/**************************************************************************\
Copyright (C) 2012-2013 Armin Biere JKU Linz

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
\**************************************************************************/

#include <assert.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

/*------------------------------------------------------------------------*/
/* In order to debug the delta-debugger define 'SHOWFILE', which will print
 * all the written temporary and reduced files.
 *
#define SHOWFILE
 */

/*------------------------------------------------------------------------*/

static char * iname, * rname, * tname, *cmdstr;
static int szinput, ninput, nreduced, savednreduced, runs;
static char * input, * reduced, * saved;
static int verbosity, round, golden;

/*------------------------------------------------------------------------*/

static struct { int chars, lines, sexprs; } only;
static struct { int chars, lines, sexprs; } enable;

/*------------------------------------------------------------------------*/

static double start;

static double currentime (void) {
  double res = 0;
  struct timeval tv;
  if (!gettimeofday (&tv, 0)) res = 1e-6 * tv.tv_usec, res += tv.tv_sec;
  return res;
}

static double getime () { return currentime () - start; }

/*------------------------------------------------------------------------*/

static double percent (double a, double b) { return b ? 100.0 * a / b : 0; }

static int reducedsize () {
  assert (nreduced <= ninput);
  return ninput - nreduced;
}

static double reduction () {
  assert (nreduced <= ninput);
  return percent (reducedsize (), ninput);
}

/*------------------------------------------------------------------------*/

static void die (const char * fmt, ...) {
  va_list ap;
  fputs ("*** ddsexpr: ", stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

static void msg (const char * fmt, ...) {
  va_list ap;
  printf ("[ddsexpr] %.1f (%d:%d:%d) ", 
          getime (), round, runs, reducedsize ());
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stdout);
}

/*------------------------------------------------------------------------*/

static void writefile (const char * name) {
  FILE * file = fopen (name, "w");
  int i, ch;
#ifdef SHOWFILE
  int lineno = 1;
  ch = '\n';
#endif
  if (!file) die ("can not write '%s'", name);
  for (i = 0; i < ninput; i++) {
    if (reduced[i]) continue;
#ifdef SHOWFILE
    if (ch == '\n') printf ("%s\t%d\t", name, lineno++);
#endif
    ch = input[i];
    fputc (ch, file);
#ifdef SHOWFILE
    fputc (ch, stdout);
#endif
  }
#ifdef SHOWFILE
  if (ch != '\n') printf ("; ACTUALLY NEW-LINE MISSING\n");
#endif
  fclose (file);
}

static int run (const char * name) {
  const char * sfx = "1>/dev/null 2>/dev/null </dev/null";
  char * mycmd = malloc (strlen (cmdstr) + strlen (name) + strlen (sfx) + 3);
  int status, res;
  sprintf (mycmd, "%s %s %s", cmdstr, name, sfx);
  status = system (mycmd);
  free (mycmd);
  res = WEXITSTATUS (status);
  runs++;
  return res;
}

static int test (const char * context) {
  int res;
  writefile (tname);
  res = run (tname);
  if (res != golden) return 0;
  msg ("writing %.0f%% %s reduced to %s", reduction (), context, rname);
  writefile (rname);
  return 1;
}

/*------------------------------------------------------------------------*/

static int isnl (int ch) { return ch=='\n'; }
static int islispcomment (int ch) { return ch == ';'; }
static int ispace (int ch) { return ch==' ' || ch=='\t' || ch=='\r'; }
static int ispaceornl (int ch) { return ispace (ch) || isnl (ch); }
static int isany (int ch) { (void) ch; return 1; }

static int nextchar (int start, int (*isthecharfun)(int)) {
  int res;
  for (res = start; res < ninput; res++)
    if (!reduced[res] && isthecharfun (input[res])) return res;
  return res;
}

static int firstchar () { return nextchar (0, isany); }

static int countchars (int (*isthecharfun)(int)) {
  int res = 0, pos;
  for (pos = 0; pos < ninput; pos++)
    if (!reduced[pos] && isthecharfun (input[pos])) res++;
  return res;
}

/*------------------------------------------------------------------------*/

static void save () { 
  memcpy (saved, reduced, ninput);
  savednreduced = nreduced;
}

static void restore () {
  nreduced = savednreduced;
  memcpy (reduced, saved, ninput);
}

static int changed () { return nreduced != savednreduced; }

/*------------------------------------------------------------------------*/

static int reduce (int pos) {
  if (pos < 0) return 0;
  if (pos >= ninput) return 0;
  if (reduced[pos]) return 0;
  assert (nreduced < ninput);
  reduced[pos] = 1;
  nreduced++;
  assert (nreduced <= ninput);
  assert (nreduced != savednreduced);
  return 1;
}

static int reducenonws (int i) {
  if (ispaceornl (input[i])) return 0;
  return reduce (i);
}

/*------------------------------------------------------------------------*/

static void stripcomments () {		// try to strip all comments at once
  static int oldnreduced = -1;
  int pos, ch, eol, lispcomments = 0;
  if (oldnreduced == nreduced) return;
  if (oldnreduced < 0) oldnreduced = 0;
  lispcomments = countchars (islispcomment);
  if (!lispcomments) return;
  msg ("trying to strip all %d LISP comments at once", lispcomments);
  save ();
  for (pos = firstchar (); pos < ninput; pos = nextchar (pos+1, isany)) {
    ch = input[pos];
    if (ch != ';') continue;
    for (eol = pos + 1; eol < ninput; eol++) {
      if (reduced[eol]) continue;
      ch = input[eol];
      if (!isnl (ch)) continue;
      break;
    }
    while (pos < eol) reduce (pos++);
  }
  assert (changed ());
  if (!test ("comment")) restore ();
  oldnreduced = nreduced;
}

static void mergews () {	// try to merge all white space at once
  static int oldnreduced = -1;
  int pos, ch, prev = '\n', prevpos = -1, update;
  if (oldnreduced == nreduced) return;
  if (oldnreduced < 0) oldnreduced = 0;
  save ();
  for (pos = firstchar (); pos < ninput; pos = nextchar (pos+1, isany)) {
    ch = input[pos];
    update = 1;
    if (ispace (ch)) {
      if (ispaceornl (prev)) reduce (pos), update = 0;
    } else if (isnl (ch)) {
      if (isnl (prev)) reduce (pos), update = 0;
      else if (ispace (prev)) {
	assert (prevpos >= 0);
	reduce (prevpos);
      }
    }
    if (!update) continue;
    prevpos = pos;
    prev = ch;
  }
  if (changed ()) {
    msg ("trying to merge all white space at once");
    if (!test ("white space")) restore ();
  }
  oldnreduced = nreduced;
}

/*------------------------------------------------------------------------*/

static int lastnewlinemissing () {
  int i;
  for (i = ninput-1; i >= 0; i--)
    if (!reduced[i]) return input[i] != '\n';
  return 1;
}

static void ddline () {			// line based delta debugger
  static int oldnreduced = -1;
  int nlines, granularity, n, l, r, remaining, removed, i, res;
  if (oldnreduced == nreduced) return;
  if (oldnreduced < 0) oldnreduced = 0;
  nlines = countchars (isnl);
  if (lastnewlinemissing ()) {
    msg ("new line after last line missing");
    nlines++;
  }
  if (!nlines) return;
  msg ("delta-debugging %d lines", nlines);
  res = 0;
  for (granularity = nlines; granularity > 0; granularity /= 2) {
    r = 0;
    for (;;) {
      l = nextchar (r, isany);
      if (l >= ninput) break;
      r = l;
      n = 0;
      do { 
	r = nextchar (r, isnl);
	if (r >= ninput) break;
	n++;
	if (++r == ninput) break;
      } while (n < granularity);

      if (granularity < nlines &&
	  ((nlines & 1) || granularity < nlines/2)) {

	save ();
	for (i = 0; i < l; i++) reduce (i);
	for (i = r; i < ninput; i++) reduce (i);

	if (changed ()) {
	  if (verbosity)
	  msg ("granularity %d "
	       "trying to remove %d chars out-side [ %d .. %d ] %.0f%%", 
	       granularity, nreduced - savednreduced, l, r,
	       percent (nreduced - savednreduced, r - l));
	  if (test ("out-side-lines")) { res = 1; break; }
	  restore ();
	}
      }

      save ();
      for (i = l; i < r; i++) reduce (i);
      if (!changed ()) continue;
      if (verbosity)
      msg ("granularity %d "
           "trying to remove %d chars in-side [ %d .. %d ] %.0f%%", 
           granularity, nreduced - savednreduced, l, r,
	   percent (nreduced - savednreduced, r - l));
      if (test ("in-side-lines")) res = 1;
      else restore ();
    }
  }
  remaining = lastnewlinemissing () + countchars (isnl);
  assert (remaining <= nlines);
  removed = nlines - remaining;
  if (removed) msg ("removed %d lines", removed);
  mergews ();
  if (!res) oldnreduced = nreduced;
}

/*------------------------------------------------------------------------*/

static int matchlp (int l, int bound) {
  int r, open = 1, ch;
  assert (0 <= l);
  assert (l < ninput);
  assert (!reduced[l]);
  assert (input[l] == '(');
  for (r = l + 1; r < bound; r++) {
    if (reduced[r]) continue;
    ch = input[r];
    if (ch == '(') open++;
    else if (ch != ')') continue;
    else if (!--open) break;
  }
  return r;
}

static int isnonsym (int ch) {
  return ispaceornl (ch) || ch == '(' || ch == ')' || ch == ';';
}

static int ddhier (int l, int r) {	// recursive hierarchical

  int lp = 0, rp = 0, open = 0, unbalanced = 0, top = 0, unreduced = 0;
  int pos, ch, nl, nr, jump, maxopen = 0, i, res;
  save ();
  res = 0;
  for (i = l; i < r; i++) reduce (i);
  if (changed ()) {
    if (test ("cut-out expressions")) return 1;
    restore ();
  }
  for (pos = l; pos < r; pos++) {
    if (reduced[pos]) continue;
    unreduced++;
    ch = input[pos];
    if (ch == '(') {
      if (!open) top++;
      lp++;
      open++;
      if (open > maxopen) maxopen = open;
      assert (open < INT_MAX);
    } else if (ch == ')') {
      rp++;
      open--;
      if (open < 0 && unbalanced < -open)
	unbalanced = -open;
    }
  }
  if (!unreduced) return 0;
  if (lp <= 0) return 0;
  if (verbosity)
  msg ("ddhier %d in (%d, %d) "
       "%d exprs, %d '(', %d ')', %d open, %d unbalanced",
       unreduced, l, r, top, lp, rp, open, unbalanced);
  open = 0;
  for (jump = 1; jump <= maxopen; jump *= 2) {
    for (pos = l; pos < r; pos++) {
      if (reduced[pos]) continue;
      ch = input[pos];
      if (ispaceornl (ch)) continue;
      nl = nr = 0;
      if (ch == '(') {
	if (++open == jump) {
	  nl = pos;
	  nr = matchlp (pos, r);
	  if (nr < r) nr++;
	  if (nl != l || nr != r) {
	    save ();
	    for (i = l; i < nl; i++) reducenonws (i);
	    for (i = nr; i < r; i++) reducenonws (i);
	    if (changed ()) {
	      if (test ("replaced expression")) res = 1;
	      else restore ();
	    }
	    if (ddhier (nl, nr)) res = 1;
	  }
	  open--;
	  pos = nr;
	}
      } else if (ch == ')') {
	if (!open--) break;
      } else if (open + 1 == jump) {
	nl = pos;
	nr = nextchar (nl, isnonsym);
	if (nl != l || nr != r) {
	  save ();
	  for (i = l; i < nl; i++) reducenonws (i);
	  for (i = nr; i < r; i++) reducenonws (i);
	  if (changed ()) {
	    if (test ("replaced symbol")) res = 1;
	    else restore ();
	  }
	}
	pos = nr;
      }
    }
  }
  return res;
}

static void ddsexpr () {
  static int oldnreduced = -1, changed;
  if (oldnreduced == nreduced) return;
  if (oldnreduced < 0) oldnreduced = 0;
  msg ("hierarchical delta-debugging of expressions");
  changed = ddhier (0, ninput);
  mergews ();
  if (changed) oldnreduced = nreduced;
}

/*------------------------------------------------------------------------*/

static void ddchar () {		// single character delta debugger

  int unreduced, remaining, removed, l, r, n, i, granularity, res;
  static int oldnreduced = -1;
  if (oldnreduced == nreduced) return;
  if (oldnreduced < 0) oldnreduced = 0;
  unreduced = countchars (isany);
  if (!unreduced) return;
  msg ("delta-debugging of %d individual characters", unreduced);
  res = 0;
  for (granularity = unreduced; granularity > 0; granularity /= 2) {
    r = 0;
    for (;;) {
      l = nextchar (r, isany);
      if (l >= ninput) break;
      r = l;
      n = 0;
      do { 
	r = nextchar (r, isany);
	if (r >= ninput) break;
	n++;
	if (++r == ninput) break;
      } while (n < granularity);

      if (granularity < unreduced &&
	  ((unreduced & 1) || granularity < unreduced/2)) {
	save ();
	for (i = 0; i < l; i++) reducenonws (i);
	for (i = r; i < ninput; i++) reducenonws (i);
	if (changed ()) {
	  if (verbosity)
	  msg ("granularity %d "
	       "trying to remove %d chars out-side [ %d .. %d ] %.0f%%", 
	       granularity, nreduced - savednreduced, l, r,
	       percent (nreduced - savednreduced, r - l));
	  if (test ("out-side-chars")) { res = 1; break; }
	  restore ();
	}
      }

      save ();
      for (i = l; i < r; i++) if (!isnl (input[i])) reduce (i);
      if (!changed ()) continue;
      if (verbosity)
      msg ("granularity %d "
           "trying to remove %d chars in-side [ %d .. %d ] %.0f%%", 
           granularity, nreduced - savednreduced, l, r,
	   percent (nreduced - savednreduced, r - l));
      if (test ("in-side-chars")) res = 1;
      else restore ();
    }
  }
  remaining = countchars (isany);
  assert (remaining <= unreduced);
  removed = unreduced - remaining;
  if (removed) msg ("removed %d chars", removed);
  mergews ();
  if (!res) oldnreduced = nreduced;
}

/*------------------------------------------------------------------------*/

static void dd () {		// main delta debugging loop
  int oldnreduced;
  stripcomments ();
  do {
    msg ("starting round %d", ++round);
    oldnreduced = nreduced;
    mergews ();
    if (enable.lines) ddline ();
    if (enable.sexprs) ddsexpr ();
    if (enable.chars) ddchar ();
  } while (oldnreduced < nreduced);
}

/*------------------------------------------------------------------------*/

static int catchedsig;
static void (*sig_int_handler)(int);
static void (*sig_segv_handler)(int);
static void (*sig_abrt_handler)(int);
static void (*sig_term_handler)(int);

static void cleanup () {
  char * name = tname;
  if (!name) return;
  tname = 0;
  unlink (name);
  free (name);
  msg ("reduced to %.0f%% in %d runs", reduction (), runs);
}

static void resetsighandlers (void) {
  (void) signal (SIGINT, sig_int_handler);
  (void) signal (SIGSEGV, sig_segv_handler);
  (void) signal (SIGABRT, sig_abrt_handler);
  (void) signal (SIGTERM, sig_term_handler);
}

static void catchsig (int sig) {
  if (!catchedsig) {
    catchedsig = 1;
    cleanup ();
  }
  resetsighandlers ();
  raise (sig);
}

static void setsighandlers (void) {
  sig_int_handler = signal (SIGINT, catchsig);
  sig_segv_handler = signal (SIGSEGV, catchsig);
  sig_abrt_handler = signal (SIGABRT, catchsig);
  sig_term_handler = signal (SIGTERM, catchsig);
}

/*------------------------------------------------------------------------*/

static const char * usage =
"usage: ddsexpr [ <option> ...] <input> <reduced> <cmd> [ <cmdargs> ... ]\n"
"\n"
"where <option> is one of the following:\n"
"\n"
"-h   print this command line summary\n"
"-v   increase verbosity level\n"
"-l   enable delta-debugging lines\n"
"-c   enable delta-debugging characters\n"
"-s   enable delta-debugging s-expressions\n"
"\n"
"If none of '-l', '-c', or '-s' is specified, all are enabled.\n"
;

int main (int argc, char ** argv) {
  FILE * file;
  int i, ch;
  start = currentime ();
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h")) {
      printf ("%s", usage);
      exit (0);
    } else if (!iname && !strcmp (argv[i], "-v")) verbosity++;
    else if (!iname && !strcmp (argv[i], "-l")) only.lines = 1;
    else if (!iname && !strcmp (argv[i], "-c")) only.chars = 1;
    else if (!iname && !strcmp (argv[i], "-s")) only.sexprs = 1;
    else if (rname) {
      int j, len = 0;
      for (j = i; j < argc; j++) len += strlen (argv[j]) + 1;
      cmdstr = strcpy (malloc (len), argv[i++]);
      while (i < argc) strcat (strcat (cmdstr, " "), argv[i++]);
    } else if (iname) rname = strdup (argv[i]);
    else iname = strdup (argv[i]);
  }
  if (!iname) die ("no <input> file given (try '-h')");
  if (!rname) die ("no <reduced> file given (try '-h')");
  if (!cmdstr) die ("no <command> given (try '-h')");
  if (only.lines + only.chars + only.sexprs == 0)
    enable.lines = enable.chars = enable.sexprs = 1;
  if (only.lines) enable.lines = 1;
  if (only.chars) enable.chars = 1;
  if (only.sexprs) enable.sexprs = 1;
  msg ("Delta Debugger for S-Exprs Version %s",  "" VERSION);
  msg ("input file name: %s", iname);
  msg ("reduced file name: %s", rname);
  msg ("command and arguments: %s", cmdstr);
  if (!(file = fopen (iname, "r"))) die ("can not read '%s'", iname);
  msg ("reading: %s", iname);
  while ((ch = getc (file)) != EOF) {
    if (ninput == szinput) {
      szinput = szinput ? 2 * szinput : 1;
      input = realloc (input, szinput);
    }
    input[ninput++] = ch;
  }
  fclose (file);
  msg ("read %d characters", ninput);
  reduced = calloc (ninput, 1);
  saved = malloc (ninput);
  msg ("writing initial reduced file: %s", rname);
  writefile (rname);
  golden = run (rname);
  msg ("golden exit code: %d", golden);
  tname = malloc (80);
  sprintf (tname, "/tmp/ddsexpr%d", getpid ());
  msg ("temporary file: %s", tname);
  setsighandlers ();
  dd ();
  free (iname), free (rname), free (cmdstr);
  free (input), free (reduced), free (saved);
  resetsighandlers ();
  cleanup ();
  return 0;
}
