/**************************************************************************\
Copyright (C) 2013 Armin Biere JKU Linz

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
\**************************************************************************/

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <signal.h>

static char * tdir, * odir, * cmd, * ddcmd;
static int verbose, force, maxrounds, dd;

static int * ignoredcodes, nignoredcodes, szignoredcodes;

static int * exitcodes, nexitcodes, szexitcodes;

static char ** paths;
static int npaths, szpaths;

static int round, bugs;

static long long seed;
static unsigned rngz, rngw;

static unsigned * cache, cached, searched, hits;

static double start;

static double currentime (void) {
  double res = 0;
  struct timeval tv;
  if (!gettimeofday (&tv, 0)) res = 1e-6 * tv.tv_usec, res += tv.tv_sec;
  return res;
}

static unsigned myrand () {
  unsigned res;
  rngz = 36969 * (rngz & 65535) + (rngz >> 16);
  rngw = 18000 * (rngw & 65535) + (rngw >> 16);
  res = (rngz << 16) + rngw;
  return res;
}

static void mysrand (unsigned long long s) {
  s ^= 0x5555555555555555;
  s *= 3111222337u;
  rngw = seed;
  rngz = seed >> 32;
  rngw *= 2019164533u;
  rngz *= 1000632769u;
}

static double getime () { return currentime () - start; }

static void msg (const char * fmt, ...) {
  va_list ap;
  printf ("[fzsexpr] %.1f %d/%d ", getime (), bugs, round);
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stdout);
}

static void wrn (const char * fmt, ...) {
  va_list ap;
  printf ("[fzsexpr] WARNING: ");
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stdout);
}

static void die (const char * fmt, ...) {
  va_list ap;
  fputs ("*** fzsexpr: ", stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

static void appcmd (char * str) {
  if (cmd) {
    int len = strlen (cmd) + strlen (str) + 2;
    cmd = realloc (cmd, len);
    strcat (cmd, " ");
    strcat (cmd, str);
  } else cmd = strdup (str);
}

static char * path (const char * name) {
  char * path = malloc (strlen (tdir) + strlen (name) + 2);
  sprintf (path, "%s/%s", tdir, name);
  return path;
}

static void pushpath (char * p) {
  if (npaths == szpaths) {
    szpaths = szpaths ? 2*szpaths : 1;
    paths = realloc (paths, szpaths * sizeof *paths);
  }
  paths[npaths++] = p;
  if (verbose) msg ("adding file: %s", p);
}

static int isfile (const char * p) {
  struct stat buf;
  if (stat (p, &buf)) return 0;
  return S_ISREG (buf.st_mode);
}

static int isdir (const char * p) {
  struct stat buf;
  if (stat (p, &buf)) return 0;
  return S_ISDIR (buf.st_mode);
}

static int existsfileordir (const char * p) {
  struct stat buf;
  return !stat (p, &buf);
}

static int findpaths (void) {
  DIR * dir = opendir (tdir);
  struct dirent * de;
  const char * name;
  char * p;
  if (!dir) die ("can not open temporary directory '%s'", tdir);
  while ((de = readdir (dir))) {
    if (!strcmp ((name = de->d_name), ".")) continue;
    if (!strcmp (name, "..")) continue;
    p = path (name);
    if (isfile (p)) pushpath (p);
    else free (p);
  }
  closedir (dir);
  msg ("found %d template files in '%s'", npaths, tdir);
  return npaths;
}

static void erase () { 
  int i;
  if (isatty (1)) {
    fputc ('\r', stdout);
    for (i = 0; i < 78; i++) fputc (' ', stdout);
    fputc ('\r', stdout);
  } else fputc ('\n', stdout);
  fflush (stdout);
}

static char * readfile (const char * p, int * lenptr) {
  FILE * file = fopen (p, "r");
  int ch, szres, nres;
  char * res;
  if (!file) return 0;
  nres = 0;
  res = malloc ((szres = 1024));
  while ((ch = getc (file)) != EOF) {
    if (nres == szres) res = realloc (res, szres *= 2);
    res[nres++] = ch;
  }
  fclose (file);
  if (nres == szres) res = realloc (res, szres *= 2);
  res[nres] = 0;
  *lenptr = nres;
  return res;
}

static unsigned primes[] =
{
  123123137u, 123123151u, 123123229u, 123123277u, 123123283u, 123123307u,
  123123311u, 123123323u, 123123337u, 123123349u, 123123359u, 123123373u,
  123123379u, 123123391u, 123123397u, 123123461u, 123123487u, 123123493u,
};

static const int nprimes = sizeof primes / sizeof *primes;

static unsigned hash (const char * t, int len) {
  unsigned res = len;
  int i, j = 0;
  for (i = 0; i < len; i++) {
    res += primes[j++] * (unsigned) t[i];
    if (j == nprimes) j = 0;
  }
  return res;
}

static int trunchars (char * t, int * lenptr, int force) {
  int oldlen = *lenptr, newlen;
  if (!oldlen) return 0;
  if (!force && (myrand () % 5)) return 0;
  newlen = myrand () % oldlen;
  if (verbose) msg ("truncated from %d bytes to %d bytes", oldlen, newlen);
  *lenptr = newlen;
  return 1;
}

static int lograndrange (int range) {
  int res, w = 1;
  while (w < range && (myrand() %  3)) w *= 2;
  if (w > range) w = range;
  res = myrand () % w;
  return res;
}

static int delchars (char * t, int * lenptr, int force) {
  int oldlen = *lenptr, newlen = oldlen;
  int n, i, j, pos, deleted;
  if (!oldlen) return 0;
  if (!force && (myrand () % 3)) return 0;
  n = lograndrange (oldlen);
  if (!n) n++;
  while (((long long)n) * ((long long) oldlen) > 1000000) n /= 2;
  deleted = 0;
  for (i = 0; newlen && i < n; i++) {
    pos = myrand () % newlen;
    for (j = pos; j + 1 < newlen; j++) t[j] = t[j+1];
    assert (j + 1 == newlen);
    t[--newlen] = 0;
    deleted++;
  }
  if (verbose) 
    msg ("deleted %d bytes from %d bytes (now %d bytes)",
         deleted, oldlen, newlen);
  *lenptr = newlen;
  return deleted;
}

static int isasciimutatable (int ch) {
  if (ch == '\r' || ch == '\n' || ch == '\t') return 1;
  return 32 <= ch && ch <= 126;
}

static int asciimutate (char * t, int * lenptr, int force) {
  int len = *lenptr, again = len, n, i, pos, mutated = 0, oldch, newch;
  if (!len) return 0;
  if (!force && (myrand () % 2)) return 0;
  n = lograndrange (len);
  if (!n) n++;
  for (i = 0; i < n; i++) {
    pos = myrand () % len;
    oldch = t[pos];
    if (!isasciimutatable (oldch)) { if (again--) i--; continue; }
    switch (myrand () % 3) {
      default: newch = (myrand () % 127); break;
      case 1: newch = oldch - 1; break;
      case 2: newch = oldch + 1; break;
    }
    if (!isasciimutatable (newch)) newch = ' ';
    t[pos] = newch;
    mutated += (oldch != newch);
  }
  if (verbose) msg ("mutated %d bytes out of %d bytes", mutated, len);
  return mutated;
}

static int flipbits (char * t, int * lenptr, int force) {
  int len = *lenptr, n, i, flipped = 0, pos, oldch, bit, newch;
  if (!len) return 0;
  if (!force && (myrand () % 5)) return 0;
  n = lograndrange (len);
  if (!n) n++;
  for (i = 0; i < n; i++) {
    pos = myrand () % len;
    oldch = t[pos];
    bit = myrand () % 8;
    newch = oldch ^ (1 << bit);
    t[pos] = newch;
    flipped++;
  }
  if (verbose) msg ("flipped %d bits in %d bytes", flipped, len);
  return flipped;
}

static int fuzz (char * t, int * lenptr) {
  int res, force = (myrand () % 4);
  res = trunchars (t, lenptr, (force == 0));
  if (delchars (t, lenptr, (force == 1))) res = 1;
  if (asciimutate (t, lenptr, (force == 2))) res = 1;
  if (flipbits (t, lenptr, (!res || (force == 3)))) res = 1;
  return res;
}

static char * createodirpath (const char * prefix, unsigned h) {
  char * res = malloc (strlen (odir) + strlen (prefix) + 10);
  sprintf (res, "%s/%s%08x", odir, prefix, h);
  return res;
}

static char * createtmppath (const char * prefix, unsigned h) {
  char * res = malloc (strlen (prefix) + 15);
  sprintf (res, "/tmp/%s%08x", prefix, h);
  return res;
}

static void writefile (const char * t, int len, const char * p) {
  FILE * file = fopen (p, "wb");
  int i;
  if (!file) die ("failed to write '%s'", p);
  for (i = 0; i < len; i++) fputc (t[i], file);
  fclose (file);
}

static void incexitcode (int code) {
  int i;
  for (i = 0; i < nexitcodes; i++)
    if (exitcodes[2*i] == code) break;
  if (i == nexitcodes) {
    szexitcodes = szexitcodes ? 2*szexitcodes : 1;
    exitcodes = realloc (exitcodes,
                         2 * szexitcodes * sizeof *exitcodes);
    exitcodes[2*i + 0] = code;
    exitcodes[2*i + 1] = 0;
    nexitcodes++;
  }
  exitcodes[2*i + 1]++;
}

static int ignoredcode (int code) {
  int i;
  for (i = 0; i < nignoredcodes; i++)
    if (ignoredcodes[i] == code) return 1;
  return 0;
}

static void copyfile (const char * src, const char * dst) {
  FILE * from = fopen (src, "rb"), * to;
  int ch;
  if (!from) die ("can not read '%s'", src);
  to = fopen (dst, "wb");
  if (!to) fclose (from), die ("can not write to '%s'", dst);
  while ((ch = fgetc (from)) != EOF) fputc (ch, to);
  fclose (from);
  fclose (to);
}

static int systemcall (const char * systemcmd) {
  int status, res;
  double time;
  if (verbose) msg ("system (\"%s\")", systemcmd);
  time = getime ();
  status = system (systemcmd);
  time = getime () - time;
  res = WEXITSTATUS (status);
  if (verbose)
    msg ("result %d (status %d) in %.3f seconds", res, status, time);
  return res;
}

static long bytesfile (const char * p) {
  struct stat buf;
  if (stat (p, &buf)) return 0;
  return buf.st_size;
}

static double percent (double a, double b) { return b ? 100.0 * a / b : 0; }

static int findcache (unsigned h) {
  int res;
  searched++;
  res = (cache[h >> 5] & (1 << (h & 31))) != 0;
  if (res) hits++;
  return res;
}

static void insertcache (unsigned h) {
  assert (!findcache (h));
  cache[h >> 5] |= (1 << (h & 31));
  cached++;
}

static void runround () {
  char * fzzname, * bugname, * redname, * systemcmd;
  const char * name;
  int i, len, res;
  const char * p;
  long bb, rb;
  unsigned h;
  char * t;
  assert (npaths > 0);
  i = (myrand () % npaths);
  p = paths[i];
  if (!p) return;
  round++;
  if ((t = readfile (p, &len))) {
    if (verbose) {
      msg ("");
      msg ("starting round %d", round);
      msg ("picked file #%d '%s'", i, p);
    } else printf ("[fzsexpr] round %d file #%d", round, i);
    if (fuzz (t, &len)) {
      h = hash (t, len);
      if (!findcache (h)) {
	insertcache (h);
	fzzname = createtmppath ("fzz", h);
	if (!verbose) printf (" %s", fzzname), fflush (stdout);
	bugname = createodirpath ("bug", h);
	redname = createodirpath ("red", h);
	if (!isfile (name = fzzname) && 
	    !isfile (name = bugname) &&
	    !isfile (name = redname)) {
	  if (verbose) msg ("writing %s", fzzname);
	  writefile (t, len, fzzname);
#if 0
{
  systemcmd = malloc (strlen (p) + strlen (fzzname) + 120);
  sprintf (systemcmd, 
    "gnome-terminal --geometry=200x80 -e \"vimdiff %s %s\"",
    p, fzzname);
  (void) systemcall (systemcmd);
  free (systemcmd);
}
#endif
	  systemcmd = malloc (strlen (cmd) + strlen (fzzname) + 120);
	  sprintf (systemcmd, 
	    "%s %s 1>/dev/null 2>/dev/null </dev/null", cmd, fzzname);
	  res = systemcall (systemcmd);
	  incexitcode (res);
	  free (systemcmd);
	  if (ignoredcode (res)) {
	    if (!verbose) erase ();
	    if (verbose) msg ("ignoring exit code %d", res);
	  } else {
	    copyfile (fzzname, bugname);
	    bugs++;
	    if (!verbose) erase ();
	    msg ("copied %s to %s", fzzname, bugname);
	    if (dd) {
	      systemcmd = malloc (strlen (ddcmd) + 
				  strlen (bugname) + 
				  strlen (redname) + 
				  strlen (cmd) + 120);
	      sprintf (systemcmd,
		"%s %s %s %s 1>/dev/null 2>/dev/null </dev/null",
		ddcmd, bugname, redname, cmd);
	      (void) systemcall (systemcmd);
	      free (systemcmd);
	      msg ("produced reduced file %s", redname);
	      bb = bytesfile (bugname);
	      rb = bytesfile (redname);
	      msg ("reduced %ld bytes to %ld bytes %.0f%%",
		   bb, rb, percent (rb, bb));
	    }
	  }

	  unlink (fzzname);
	} else {
	  if (!verbose) erase ();
	  assert (name);
	  wrn ("file '%s' already exists", name);
	}

	free (fzzname);
	free (redname);
	free (bugname);

      } else if (verbose) msg ("file hash cached so ignored");
      else erase ();
    } else if (verbose) msg ("fuzzing without change for round %d file #%d",
                             round, i);
    else erase ();

    free (t);

  } else {
    wrn ("can not read file #%d '%s'", i, p);
    free (paths[i]), paths[i] = 0;
  }
}

static const char * usage = 
"usage: fzsexprs [ <opt> ... ] <tdir> <odir> <cmd> [ <arg> ... ]\n"
"\n"
"where <opt>   is one of the following options:\n"
"\n"
"  -h          print this command line option summary\n"
"  -v          increase verbosity\n"
"  -f          force overwriting\n"
"\n"          
"  -s <seed>   random number generator seed (64 bit)\n"
"  -m <rounds> limit number of runs (default unlimited)\n"
"  -i <code>   ignore this exit code\n"
"  -d <dd>     use this delta debugging command (default 'ddsexprs')\n"
"\n"          
"  --sat       ignore exit code '0', '1', '10' and '20' (default)\n"
"  --plain     ignore only exit code '0'\n"
"  --no-dd     do not delta debug\n"
"\n"
"The template directory is specified as '<tdir>', the output directory\n"
"as '<odir>', and the command to be checked as '<cmd>'.  Additional\n"
"command line arguments for the command follow at the end.\n"
;

// Not really full normalization, but just standard (Unix) cases,
// e.g. strip leading './' and trailing '/'.

static char * normalizedirname (const char * name) {
  char * res;
  int len;
  if (name[0] == '.' && name[1] == '/' && name[2]) res = strdup (name + 2);
  else res = strdup (name);
  while ((len = strlen (res)) > 1 && res[len-1] == '/') res[len-1] = 0;
  return res;
}

static double average (double a, double b) { return b ? a / b : 0; }

static int cmpexitcodes (const void * p, const void * q) {
  const int * c = (const int *) p, * d = (const int*) q;
  int res = d[1] - c[1];
  if (res) return res;
  return c[0] - d[0];
}

static void stats () {
  double time = getime ();
  int i;
  msg ("exit code distribution:");
  msg ("");
  qsort (exitcodes, nexitcodes, 2*sizeof *exitcodes, cmpexitcodes);
  for (i = 0; i < nexitcodes; i++)
    msg ("  exit code %-3d returned %8d times %3.0f%%",
	 exitcodes[2*i], exitcodes[2*i+1],
	 percent (exitcodes[2*i+1], round));
  msg ("");
  msg ("finished %d rounds in %.0f sec (%.2f rounds/sec)",
       round, time);
  msg ("fuzzing rate of %.1f rounds/sec", average (round, time));
  msg ("inserted %d file hashs into cache %.3f%% filled", 
       cached, percent (cached, (1ll << 32)));
  msg ("searched %d, hits %d, hit rate %.3f%%",
       searched, hits, percent (hits, searched));
  msg ("found %d bugs, %.1f sec/bug, %.2f%% rounds/bug",
    bugs, average (time, bugs), percent (round, bugs));
}

static int catchedsig;

static void (*sig_int_handler)(int);
static void (*sig_segv_handler)(int);
static void (*sig_abrt_handler)(int);
static void (*sig_term_handler)(int);

static void resetsighandlers (void) {
  (void) signal (SIGINT, sig_int_handler);
  (void) signal (SIGSEGV, sig_segv_handler);
  (void) signal (SIGABRT, sig_abrt_handler);
  (void) signal (SIGTERM, sig_term_handler);
}

static void catchsig (int sig) {
  if (!catchedsig) {
    catchedsig = 1;
    stats ();
  }
  resetsighandlers ();
  raise (sig);
}

static void setsighandlers (void) {
  sig_int_handler = signal (SIGINT, catchsig);
  sig_segv_handler = signal (SIGSEGV, catchsig);
  sig_abrt_handler = signal (SIGABRT, catchsig);
  sig_term_handler = signal (SIGTERM, catchsig);
}

static void ignorecode (int code) {
  if (ignoredcode (code)) return;
  if (szignoredcodes == nignoredcodes) {
    szignoredcodes = szignoredcodes ? 2*szignoredcodes : 1;
    ignoredcodes = 
      realloc (ignoredcodes, szignoredcodes * sizeof *ignoredcodes);
  }
  ignoredcodes[nignoredcodes++] = code;
}

static void ignoreplain () {
  nignoredcodes = 0;
  ignorecode (0);
}

static void ignoresat () {
  nignoredcodes = 0;
  ignorecode (0);
  ignorecode (1);
  ignorecode (10);
  ignorecode (20);
}

static void cleanup () {
  int i;
  free (tdir), free (odir), free (cmd), free (ddcmd);
  free (ignoredcodes);
  for (i = 0; i < npaths; i++) free (paths[i]);
  free (paths);
  free (exitcodes);
  free (cache);
}

int main (int argc, char ** argv) {
  int i, code;
  start = currentime ();
  maxrounds = -1, seed = -1;
  ignoresat ();
  ddcmd = strdup ("ddsexpr"), dd = 1;
  for (i = 1; i < argc; i++) {
    if (!tdir && !strcmp (argv[i], "-v")) verbose++;
    else if (!tdir && !strcmp (argv[i], "-f")) force++;
    else if (!tdir && !strcmp (argv[i], "-s")) {
      if (++i >= argc) die ("argument to '-s' missing");
      if ((seed = atoll (argv[i])) < 0)
	die ("invalid seed argument in '-s %s'", argv[i]);
    } else if (!tdir && !strcmp (argv[i], "-m")) {
      if (++i >= argc) die ("argument to '-m' missing");
      if ((maxrounds = atoi (argv[i])) < 0)
	die ("invalid runs argument in '-m %s'", argv[i]);
    } else if (!tdir && !strcmp (argv[i], "-i")) {
      if (++i >= argc) die ("argument to '-i' missing");
      if ((code = atoi (argv[i])) < 0)
	die ("invalid code in '-i %s'", argv[i]);
      ignorecode (code);
    } else if (!tdir && !strcmp (argv[i], "-d")) {
      if (++i >= argc) die ("argument to '-d' missing");
      free (ddcmd);
      ddcmd = strdup (argv[i]);
    } else if (!tdir && !strcmp (argv[i], "--sat")) ignoresat ();
    else if (!tdir && !strcmp (argv[i], "--plain")) ignoreplain ();
    else if (!tdir && !strcmp (argv[i], "--no-dd")) dd = 0;
    else if (!tdir && !strcmp (argv[i], "-h")) {
      printf ("%s", usage);
      exit (0);
    } else if (!tdir && argv[i][0] == '-')
      die ("invalid command line option '%s' (try '-h')", argv[i]);
    else if (odir) appcmd (argv[i]);
    else if (tdir) odir = normalizedirname (argv[i]);
    else tdir = normalizedirname (argv[i]);
  }
  if (!tdir) die ("template directory missing (try '-h')");
  if (!isdir (tdir)) die ("template argument '%s' not a directory", tdir);
  if (!odir) die ("output directory missing (try '-h')");
  if (existsfileordir (odir)) {
     if (!force) 
       die ("not overwriting existing output directory '%s' (use '-f')", odir);
     if (!isdir (odir))
       die ("output argument '%s' exists but is not a directory", odir);
  }
  if (!cmd) die ("command argument missing (try '-h')");
  if (!strcmp (tdir, odir)) {
    if (force) wrn ("forced to use identical template and output directory");
    else die ("template and output directory identical (use '-f')");
  }
  msg ("Fuzzer for S-Exprs Version %s", VERSION);
  msg ("template directory '%s'", tdir);
  if (!isdir (odir)) {
    if (mkdir (odir, 0777))
      die ("failed to create output directory '%s'", odir);
    msg ("made output directory '%s'", odir);
  } else msg ("reusing existing output directory '%s'", odir);
  msg ("command '%s'", cmd);
  if (!findpaths ()) die ("no template files found in '%s'", tdir);
  if (seed < 0) {
    seed = getpid ();
    seed <<= 31;
    seed |= times(0);
    assert (seed >= 0);
    msg ("automatically generated seed %lld", seed);
  } else msg ("using specified seed %lld", seed);
  mysrand (seed);
  for (i = 0; i < nignoredcodes; i++)
    msg ("will ignore exit code %d", ignoredcodes[i]);
  if (dd) msg ("using delta debugging command '%s'", ddcmd);
  else msg ("delta debugging disabled");
  cache = (unsigned *) malloc (1 << 29);
  msg ("allocated 512 MB cache");
  memset (cache, 0, (1 << 29));
  msg ("initialized 512 MB cache");
  if (maxrounds < 0) msg ("infinitely many rounds (use '-m ...')");
  else msg ("number of rounds limited to %d as specified", maxrounds);
  setsighandlers ();
  while ((maxrounds < 0) || round < maxrounds)
    runround ();
  if (verbose && round) msg ("");
  resetsighandlers ();
  stats ();
  cleanup ();
  return 0;
}
